from AIBridge.ai_services.cohere_llm import CohereApi
from AIBridge.ai_services.palm_chat import PalmChat
from AIBridge.ai_services.palm_text import PalmText
from AIBridge.queue_integration.response_class import (
    OpenAiImageRes,
    PalmTextRes,
    OpenAiRes,
    PalmChatRes,
    StableDuffusionRes,
    CohereRes,
    JarasicTextRes,
)
from AIBridge.ai_services.openai_images import OpenAIImage
from AIBridge.ai_services.stable_diffusion_image import StableDiffusion
from AIBridge.exceptions import ProcessMQException
from AIBridge.ai_services.ai21labs_text import JurasicText


class ProcessMQ:
    @classmethod
    def get_process_mq(self, process_name):
        from AIBridge.ai_services.openai_services import OpenAIService

        process_obj = {
            "open_ai": OpenAIService(),
            "open_ai_image": OpenAIImage(),
            "palm_api": PalmText(),
            "palm_chat": PalmChat(),
            "stable_diffusion": StableDiffusion(),
            "cohere_api": CohereApi(),
            "ai21_api": JurasicText(),
        }
        response_obj = {
            "open_ai": OpenAiRes(),
            "open_ai_image": OpenAiImageRes(),
            "palm_api": PalmTextRes(),
            "palm_chat": PalmChatRes(),
            "stable_diffusion": StableDuffusionRes(),
            "cohere_api": CohereRes(),
            "ai21_api": JarasicTextRes(),
        }
        if process_name not in process_obj:
            raise ProcessMQException(
                f"Process of message queue Not Found process->{process_name}"
            )
        return process_obj[process_name], response_obj[process_name]



import sqlite3

from AIBridge.exceptions import VaribalesException
from AIBridge.database.models.prompts import Prompts, Base
from AIBridge.exceptions import AIBridgeException, PromptSaveException
from AIBridge.database.models.variables import Variables, Base as v_base
from AIBridge.database.db_layer import DBLayer
import uuid
import json
import calendar
import time


class PromptInsertion:
    @classmethod
    def get_time(self):
        return calendar.timegm(time.gmtime())

    @classmethod
    def validate_variables(self, variables):
        for key, value in variables.items():
            variable = DBLayer.filter_table(Variables, **{"key": value})
            if not variable:
                raise VaribalesException(f"Invalid variable key->{key}")

    @classmethod
    def validate_prompt(
        self, prompt=None, name=None, prompt_data={}, variables={}, id=None
    ):
        if id:
            try:
                uuid.UUID(id)
            except AIBridgeException as e:
                raise PromptSaveException("id must be a valid uuid")
        if prompt:
            if not isinstance(prompt, str):
                raise PromptSaveException("prompt must be a string")
        if not isinstance(prompt_data, dict):
            raise PromptSaveException("prompt_data must be a json")
        if not isinstance(variables, dict):
            raise PromptSaveException("variables must be a json")
        if name:
            if not isinstance(name, str):
                raise PromptSaveException("name must be a string")

    @classmethod
    def save_prompt(self, prompt: str, name: str, prompt_data={}, variables={}):
        # Validate parameters
        self.validate_prompt(
            prompt=prompt, name=name, prompt_data=prompt_data, variables=variables
        )
        method = "save"
        table_exist = DBLayer.check_table(Prompts, Base, method)
        prompt_n = DBLayer.filter_table(Prompts, **{"name": name})
        if prompt_n:
            raise PromptSaveException(f"prompt already exists with the name->{name}")
        if variables:
            self.validate_variables(variables)
        prompt_obj = {
            "id": str(uuid.uuid4()),
            "name": name,
            "prompt": prompt,
            "prompt_data": json.dumps(prompt_data),
            "variables": json.dumps(variables),
            "updated_at": self.get_time(),
            "created_at": self.get_time(),
        }
        return DBLayer.save(Prompts, Base, prompt_obj)

    @classmethod
    def update_prompt(
        self, id: str, prompt=None, name=None, prompt_data={}, variables={}
    ):
        self.validate_prompt(
            prompt=prompt if prompt else None,
            name=name if name else None,
            prompt_data=prompt_data,
            variables=variables,
            id=id,
        )
        prompt_obj = DBLayer.get_by_id(Prompts, id)
        if name:
            result = DBLayer.filter_table(Prompts, **{"name": name})
            if result:
                if result["name"] != name:
                    raise PromptSaveException(
                        f"prompt already exists with the name->{name}"
                    )
        prompt_schema = {
            "name": name if name else None,
            "prompt": prompt if prompt else None,
            "prompt_data": json.dumps(prompt_data) if prompt_data else None,
            "variables": json.dumps(variables) if variables else None,
            "updated_at": self.get_time(),
        }
        return DBLayer.update(Prompts, prompt_schema, id)

    @classmethod
    def get_prompt(self, id):
        self.validate_prompt(id=id)
        return DBLayer.get_by_id(Prompts, id)

    @classmethod
    def get_all_prompt(self, page):
        return DBLayer.get_all(Prompts, page)



from AIBridge.constant.constant import CUSTOM_CONFIG_PATH, DB_TYPES, LINUX_MACOS_PATH
import yaml
from AIBridge.constant.constant import PRIORITY, AI_SERVICES
from AIBridge.exceptions import ConfigException
import os


def get_config_path():
    path = "aibridge_config.yaml"
    if os.environ.get(CUSTOM_CONFIG_PATH):
        path = os.environ.get(CUSTOM_CONFIG_PATH)
        directory = os.path.dirname(path)
        os.makedirs(directory, exist_ok=True)
    return path


class SetConfig:
    @classmethod
    def read_yaml(self):
        try:

            def get_yaml():
                path = get_config_path()
                with open(path, "r") as file:
                    data = yaml.safe_load(file)
                    return data

            return get_yaml()
        except FileNotFoundError as e:
            self.write_yaml({})
            return get_yaml()

    @classmethod
    def write_yaml(self, data):
        path = get_config_path()
        with open(path, "w") as file:
            yaml.safe_dump(data, file)

    @classmethod
    def set_api_key(self, ai_service, key, priority="equal"):
        if ai_service not in AI_SERVICES:
            raise ConfigException(
                "key for ai service is must be one of these: {}".format(AI_SERVICES)
            )
        if priority not in PRIORITY:
            raise ConfigException(
                "priority for ai service is must be one of these: {}".format(PRIORITY)
            )
        data = self.read_yaml()
        if ai_service in data:
            if priority in data[ai_service]:
                data[ai_service][priority].append(key)
            else:
                data[ai_service][priority] = [key]
        else:
            data[ai_service] = {priority: [key]}
        self.write_yaml(data)

    @classmethod
    def redis_config(
        self,
        redis_host,
        redis_port,
        group_name,
        stream_name,
        no_of_threads=1,
    ):
        data = self.read_yaml()
        if not data:
            data = {}
        data["message_queue"] = "redis"
        for arg_name, arg_value in locals().items():
            if arg_name == "self":
                continue
            if arg_value:
                data[arg_name] = arg_value
        self.write_yaml(data)

    @classmethod
    def set_db_confonfig(self, database="sql", database_name=None, database_uri=None):
        if database not in DB_TYPES:
            raise ConfigException(
                "key for database service is must be one of these: {}".format(DB_TYPES)
            )
        data = self.read_yaml()
        if not data:
            data = {}
        data["database"] = database
        if database_name:
            data["database_name"] = database_name
        if database_uri:
            data["database_uri"] = database_uri
        self.write_yaml(data)



from AIBridge.output_validation.validations import JsonSchema, SQLSchema, XMLSchema, CSVSchema


class ActiveValidator:
    @classmethod
    def get_active_validator(self, format):
        validation_dict = {
            "json": JsonSchema(),
            "sql": SQLSchema(),
            "xml": XMLSchema(),
            "csv": CSVSchema(),
        }
        return validation_dict[format]



from AIBridge.ai_services.openai_services import OpenAIService
from AIBridge.exceptions import ConfigException, OpenAIException, PromptSaveException
from AIBridge.setconfig import SetConfig
import AIBridge.exceptions as exceptions
from AIBridge.prompts.prompts_save import PromptInsertion
from AIBridge.prompts.prompts_varibales import VariableInsertion
from AIBridge.ai_services.ai_services_response import FetchAIResponse
from AIBridge.queue_integration.message_queue import MessageQ
from AIBridge.output_validation.active_validator import ActiveValidator
from AIBridge.output_validation.validations import Validation
from AIBridge.ai_services.openai_images import OpenAIImage
from AIBridge.ai_services.palm_text import PalmText
from AIBridge.ai_services.palm_chat import PalmChat
from AIBridge.ai_services.stable_diffusion_image import StableDiffusion
from AIBridge.ai_services.cohere_llm import CohereApi
from AIBridge.ai_services.ai21labs_text import JurasicText


__all__ = [
    "OpenAIService",
    "SetConfig",
    "COnfigException",
    "OpenAIException",
    "PromptSaveException",
    "VariableInsertion",
    "PromptInsertion",
    "FetchAIResponse",
    "MessageQ",
    "ActiveValidator",
    "Validation",
    "OpenAIImage",
    "PalmText",
    "PalmChat",
    "StableDiffusion",
    "CohereApi",
    "JurasicText",
]

__version__ = "0.0.0"



"""
Exception  handling
"""


class AIBridgeException(Exception):
    pass


class OpenAIException(AIBridgeException):
    pass


class ConfigException(AIBridgeException):
    pass


class ValidationException(AIBridgeException):
    pass


class PromptSaveException(AIBridgeException):
    pass


class VaribalesException(AIBridgeException):
    pass


class PromptCompletionException(AIBridgeException):
    pass


class AssignQueueException(AIBridgeException):
    pass


class ProcessMQException(AIBridgeException):
    pass


class AIResponseException(AIBridgeException):
    pass


class MessageQueueException(AIBridgeException):
    pass


class DatabaseException(AIBridgeException):
    pass


class ImageException(AIBridgeException):
    pass


class PalmTextException(AIBridgeException):
    pass


class StableDiffusionException(AIBridgeException):
    pass


class CohereException(AIBridgeException):
    pass


class Ai21Exception(AIBridgeException):
    pass



from AIBridge.constant.constant import FUNCTION_CALL_FORMAT, PRIORITY
from AIBridge.setconfig import SetConfig
from AIBridge.exceptions import ConfigException
from AIBridge.database.sql_service import SQL
from AIBridge.database.no_sql_service import Mongodb
from AIBridge.exceptions import AIBridgeException
from urllib.parse import urlparse
from AIBridge.output_validation.convertors import FromJson, IntoJson
import json

config = SetConfig.read_yaml()


def parse_fromat(prompt, format=None, format_structure=None):
    if format:
        prompt = prompt + f"format:json valid"
    if format_structure:
        if format:
            if format == "csv":
                format_structure = json.dumps(IntoJson.csv_to_json(format_structure))
            elif format == "xml":
                format_structure = json.dumps(IntoJson.xml_to_json(format_structure))

        prompt = prompt + f"format_structure:{format_structure}"
    prompt = (
        prompt
        + "Respond only in the exact specified format provided in the prompt,No extra information,No extra space"
    )
    return prompt


def parse_api_key(ai_service):
    if ai_service not in config:
        raise ConfigException("ai_service not found in config file")
    return config[ai_service][0]["key"]


def get_no_sql_obj():
    databse_uri = config["database_uri"]
    if "mongodb" in databse_uri:
        return Mongodb()


def get_database_obj():
    if "database" not in config:
        return SQL()
    elif config["database"] == "nosql":
        return get_no_sql_obj()
    elif config["database"] == "sql":
        return SQL()


def get_function_from_json(output_schema: dict):
    def create_function_call(output_schema: dict):
        key_dict = {"type": "object", "properties": {}}
        for key, value in output_schema.items():
            if isinstance(value, dict):
                key_dict["properties"][key] = create_function_call(value)
            elif isinstance(value, list):
                key_d = {
                    "type": "array",
                    "description": "Generate the information",
                }
                if value:
                    if isinstance(value[0], str):
                        key_d["items"] = {"type": "string", "description": value[0]}
                    elif isinstance(value[0], dict):
                        key_d["items"] = create_function_call(value[0])
                else:
                    key_d["items"] = {
                        "type": "string",
                        "description": "provide the given information",
                    }
                key_dict["properties"][key] = key_d
            else:
                key_dict["properties"][key] = {"type": "string", "description": value}
        return key_dict

    required = []
    for key in output_schema.keys():
        required.append(key)
    data = FUNCTION_CALL_FORMAT
    key_dict = create_function_call(output_schema)
    data["parameters"] = key_dict
    data["parameters"]["required"] = required
    return data


def check_url(url_list: list):
    for url in url_list:
        result = urlparse(url)
        if not all([result.scheme, result.netloc]):
            raise AIBridgeException(
                "Only image url is required, please provide valid one"
            )
    return



from AIBridge.prompts.prompt_completion import Completion
from AIBridge.ai_services.ai_abstraction import AIInterface
from AIBridge.exceptions import OpenAIException, AIBridgeException
import json
import uuid
from AIBridge.constant.constant import OPENAI_IMAGE_TYPE, OPENAI_IMAGE_SIZES
from openai import OpenAI
from AIBridge.constant.common import parse_api_key
from AIBridge.ai_services.image_optimisaton import ImageOptimise
import time


class OpenAIImage(AIInterface):
    @classmethod
    def generate(
        self,
        prompts: list[str] = [],
        prompt_ids: list[str] = [],
        prompt_data: list[dict] = [],
        variables: list[dict] = [],
        image_data: list[str] = [],
        mask_image: list[str] = [],
        variation_count: int = 1,
        size: str = "1024x1024",
        process_type: str = "create",
        message_queue=False,
        api_key=None,
        model="dall-e-2",
        quality="standard",
    ):
        try:
            if prompts and prompt_ids:
                raise OpenAIException(
                    "please provide either prompts or prompts ids at atime"
                )
            if not prompts and not prompt_ids:
                raise OpenAIException(
                    "Either provide prompts or prompts ids to genrate the data"
                )
            if process_type not in OPENAI_IMAGE_TYPE:
                raise OpenAIException(
                    "process_type can be either create, variation, edit"
                )
            if size not in OPENAI_IMAGE_SIZES:
                raise OpenAIException(
                    "size can be either 1024x1024 or 512x512 or 256x256"
                )
            if process_type == "edit" or process_type == "variation":
                if model != "dall-e-2":
                    raise OpenAIException(
                        "model can be only dall-e-2 for edit and variation"
                    )
                if not image_data:
                    raise OpenAIException("Please enter image for edit or variation")
            if mask_image:
                if len(mask_image) != len(image_data):
                    raise OpenAIException(
                        "mask_image length should be equal to image_data length",
                    )
            if prompt_ids:
                prompts_list = Completion.create_prompt_from_id(
                    prompt_ids=prompt_ids,
                    prompt_data_list=prompt_data,
                    variables_list=variables,
                )
            if prompts:
                if prompt_data or variables:
                    prompts_list = Completion.create_prompt(
                        prompt_list=prompts,
                        prompt_data_list=prompt_data,
                        variables_list=variables,
                    )
                else:
                    prompts_list = prompts
            if image_data:
                if prompts_list:
                    if len(image_data) != len(prompts_list):
                        raise OpenAIException(
                            "image_data length should be equal to prompts length",
                        )
            if message_queue:
                id = uuid.uuid4()
                message_data = {
                    "id": str(id),
                    "prompts": json.dumps(prompts_list),
                    "variation_count": variation_count,
                    "ai_service": "open_ai_image",
                    "image_data": json.dumps(image_data),
                    "mask_image": json.dumps(mask_image),
                    "size": size,
                    "process_type": process_type,
                    "api_key": api_key,
                    "model": model,
                    "quality": quality,
                }
                message = {"data": json.dumps(message_data)}
                from AIBridge.queue_integration.message_queue import MessageQ

                MessageQ.mq_enque(message=message)
                return {"response_id": str(id)}
            return self.get_response(
                prompts=prompts_list,
                image_data=image_data,
                mask_image=mask_image,
                variation_count=variation_count,
                size=size,
                process_type=process_type,
                api_key=api_key,
                model=model,
                quality=quality,
            )
        except AIBridgeException as e:
            raise OpenAIException(f"{e}")

    @classmethod
    def get_response(
        self,
        prompts,
        image_data=[],
        mask_image=[],
        variation_count=1,
        size="1024*1024",
        process_type="create",
        api_key=None,
        model="dall-e-2",
        quality="standard",
    ):
        try:
            OPEN_AI_API_KEY = api_key if api_key else parse_api_key("open_ai")
            client = OpenAI(api_key=api_key)
            data = []
            if image_data:
                image_data = ImageOptimise.get_image(image_data)
                if mask_image:
                    mask_image = ImageOptimise.get_image(mask_image)
                    mask_image = ImageOptimise.set_dimension(image_data, mask_image)
                    mask_image = ImageOptimise.get_bytes_io(mask_image)
                image_data = ImageOptimise.get_bytes_io(image_data)
            if process_type == "create":
                for index, prompt in enumerate(prompts):
                    response = client.images.generate(
                        model=model,
                        prompt=prompt,
                        n=variation_count,
                        size=size,
                        quality=quality,
                    )
                    _data = [
                        json.dumps({"url": obj["url"]}) for obj in response["data"]
                    ]
                    data.append({"data": _data})
            elif process_type == "edit":
                for index, prompt in enumerate(prompts):
                    image = image_data[index]
                    mask = None
                    if mask_image:
                        mask = mask_image[index]
                    response = client.images.edit(
                        image=image,
                        mask=mask,
                        prompt=prompt,
                        n=variation_count,
                        size=size,
                    )
                    _data = [
                        json.dumps({"url": obj["url"]}) for obj in response["data"]
                    ]
                    data.append({"data": _data})
            elif process_type == "variation":
                for index, image in enumerate(image_data):
                    response = client.images.create_variation(
                        image=image, n=variation_count, size=size
                    )
                    _data = [
                        json.dumps({"url": obj["url"]}) for obj in response["data"]
                    ]
                    data.append({"data": _data})
            message_value = {
                "items": {
                    "response": data,
                    "token_used": 0,
                    "created_at": time.time(),
                    "ai_service": "open_ai",
                }
            }
            return message_value
        except AIBridgeException as e:
            raise OpenAIException(f"Error in creating image using open ai{e}")



from abc import ABC, abstractmethod
import json
from AIBridge.exceptions import AIBridgeException, ValidationException
import csv
import io
import xml.etree.ElementTree as ET
import re
import sqlparse
from json import JSONDecodeError
import jsonschema


class Validation(ABC):
    @abstractmethod
    def validate(self, output_string, schema):
        pass


class JsonSchema(Validation):
    def validate(self, output_string, schema=None):
        try:
            if type(output_string) == dict:
                json_data = json.loads(output_string)
            else:
                if "}" not in output_string:
                    output_string += '"}'
                    output_string = output_string + "}"
                json_string = output_string[
                    output_string.index("{") : len(output_string)
                    - output_string[::-1].index("}")
                ]
                json_string = json_string.replace("\n", "")
                json_data = json.loads(json_string)
        except json.JSONDecodeError as e:
            raise AIBridgeException(f" Error in the AI output for the validation->{e}")
        if schema:
            try:
                user_schema = json.loads(schema)
            except AIBridgeException as e:
                raise JSONDecodeError(f"Error in the schema you entred {e}")
            try:
                jsonschema.Draft7Validator.check_schema(user_schema)
                jsonschema.Draft7Validator.check_schema(json_data)
            except jsonschema.exceptions.SchemaError as e:
                raise ValidationException(f"Invalid output JSON schema: {e}")
        return json.dumps(json_data)


class CSVSchema(Validation):
    def validate(self, output_string, schema=None):
        try:
            output_schema=[]
            if schema:
                output_schema=schema.split(",")
                if  not output_schema:
                    output_schema=schema.split(" ")
                if not output_schema:
                    raise ValidationException("CSV Schema is not in proper format ")
            reader = csv.reader(output_string.splitlines(), delimiter=",")
            new_rows=[]
            for row in reader:
                if len(row)>= len(output_schema)-1:
                    new_rows.append(row)
            if not new_rows:
                raise ValidationException("No csv data found in the output")
            rows_with_index = [(str(index), *row) for index, row in enumerate(new_rows)]
            rows_as_strings = [",".join(row) for row in rows_with_index]
            if "," not in output_schema:
                schema = ",".join(output_schema)
            result_text = "\n".join([schema] + rows_as_strings)
            result_text=result_text.replace("|",",")
            return result_text
        except csv.Error as e:
            raise ValidationException(f"{e}")


class SQLSchema(Validation):
    def validate(self, output_string, schema=None):
        sql_keywords = [
            "SELECT",
            "UPDATE",
            "INSERT",
            "DELETE",
            "FROM",
            "WHERE",
            "JOIN",
        ]
        pattern = r"\b(?:{})\b".format("|".join(sql_keywords))
        if re.search(pattern, output_string, re.IGNORECASE):
            parsed = sqlparse.parse(output_string)
            if any(parsed):
                return output_string
            else:
                raise ValidationException(
                    f"Sql schema genrated by AI is not a valid schema"
                )
        else:
            raise ValidationException(
                f"Sql schema genrated by AI is not a valid schema"
            )


class XMLSchema(Validation):
    def validate(self, output_string, schema=None):
        if """<?xml version="1.0" encoding="UTF-8"?>""" in output_string:
            output_string = output_string.replace(
                """<?xml version="1.0" encoding="UTF-8"?>""", ""
            )
        xml_match = re.search(r'```xml(?:.*?<?xml.*?>)?(.*?)```', output_string, re.DOTALL)
        if xml_match:
            output_string = xml_match.group(1)
        try:
            output_schema = ET.fromstring(output_string)
        except ET.ParseError as e:
            raise ValidationException(
                f"Xml schema generated by AI is not a valid schema->{e}"
            )
        if schema:
            input_schema = ET.fromstring(schema)

            def validate_xml(output_schema, input_schema):
                if output_schema.tag != input_schema.tag:
                    return False
                if output_schema.attrib != input_schema.attrib:
                    return False
                for child1, child2 in zip(output_schema, input_schema):
                    if not validate_xml(child1, child2):
                        return False

                return True

            result = validate_xml(output_schema, input_schema)
            if not result:
                raise ValidationException(
                    f"Xml schema generated by AI is not a valid schema"
                )
        if """<?xml version="1.0" encoding="UTF-8"?>""" not in output_string:
            output_string = """<?xml version="1.0" encoding="UTF-8"?>""" +"\n"+ output_string
        return output_string



from sqlalchemy import MetaData
from AIBridge.database.session import engine


def check_table_exist(table_name):
    connection = engine.connect()
    metadata = MetaData()
    metadata.reflect(bind=engine)
    table_exists = table_name in metadata.tables
    connection.close()
    if table_exists:
        return True
    else:
        return False


def create_table(Base):
    Base.metadata.create_all(engine)



import sqlite3
from AIBridge.exceptions import AIBridgeException, VaribalesException
from AIBridge.database.models.variables import Variables, Base
import uuid
import json
import calendar
import time
from AIBridge.database.db_layer import DBLayer


class VariableInsertion:
    @classmethod
    def get_time(self):
        return calendar.timegm(time.gmtime())

    @classmethod
    def validates_variables(self, var_key=None, var_value=None, id=None):
        if id:
            try:
                uuid.UUID(id)
            except AIBridgeException as e:
                raise VaribalesException("Invalid variable id")
        if var_key:
            if not isinstance(var_key, str):
                raise VaribalesException("Invalid variable key, it must be string")
        if var_value:
            if not isinstance(var_value, list):
                raise VaribalesException("Invalid variable value it must be list")

    @classmethod
    def save_variables(self, var_key: str, var_value: list):
        self.validates_variables(var_key=var_key, var_value=var_value)
        table_exist = DBLayer.check_table(Variables, Base, method="save")
        variable_n = DBLayer.filter_table(Variables, **{"key": var_key})
        if variable_n:
            raise VaribalesException(
                f"variable already exists with the key ->{var_key}"
            )
        str_id = str(uuid.uuid4())
        variable = {
            "id": str_id,
            "key": var_key,
            "value": json.dumps(var_value),
            "updated_at": self.get_time(),
            "created_at": self.get_time(),
        }
        return DBLayer.save(Variables, Base, data=variable)

    @classmethod
    def update_variables(self, id: str, var_key=None, var_value=None):
        self.validates_variables(id=id, var_key=var_key, var_value=var_value)
        variable = DBLayer.get_by_id(Variables, id)
        if var_key:
            variable_key = DBLayer.filter_table(Variables, **{"key": var_key})
            if variable_key:
                if variable_key["key"] != var_key:
                    raise VaribalesException(
                        f"variable already exists with the key ->{var_key}"
                    )
        updates = {
            "key": var_key,
            "value": json.dumps(var_value),
            "updated_at": self.get_time(),
        }
        return DBLayer.update(Variables, updates, id)

    @classmethod
    def get_variable(self, id):
        self.validates_variables(id=id)
        return DBLayer.get_by_id(Variables, id)

    @classmethod
    def get_all_variable(self, page=1):
        return DBLayer.get_all(Variables, page)



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from AIBridge.setconfig import SetConfig

config = SetConfig.read_yaml()


def get_uri():
    db_uri = (
        config["database_uri"]
        if "database_uri" in config
        else "sqlite:///aibridge_database.db"
    )
    if "database" in config:
        if config["database"] == "nosql":
            db_uri = "sqlite:///aibridge_database.db"
    return db_uri


engine = create_engine(get_uri(), echo=True)


def db_session():
    Base = declarative_base()
    Base.metadata.create_all(bind=engine)
    Session = sessionmaker(bind=engine)
    session = Session()
    return session



import time
import uuid
from AIBridge.exceptions import PalmTextException, AIBridgeException
from AIBridge.ai_services.ai_abstraction import AIInterface
import json
from AIBridge.constant.common import parse_fromat, parse_api_key
import google.generativeai as palm


class PalmChat(AIInterface):
    """
    Base class for PalmChat's Services
    """

    @classmethod
    def generate(
        self,
        messages: str,
        context: str = "",
        examples: list[str] = [],
        model="models/chat-bison-001",
        variation_count: int = 1,
        temperature: float = 0.5,
        message_queue=False,
        api_key=None,
    ):
        try:
            if message_queue:
                id = uuid.uuid4()
                message_data = {
                    "id": str(id),
                    "messages": messages,
                    "context": context,
                    "examples": examples,
                    "model": model,
                    "variation_count": variation_count,
                    "temperature": temperature,
                    "ai_service": "palm_chat",
                    "api_key": api_key,
                }
                message = {"data": json.dumps(message_data)}
                from AIBridge.queue_integration.message_queue import MessageQ

                MessageQ.mq_enque(message=message)
                return {"response_id": str(id)}
            return self.get_response(
                messages,
                context,
                examples,
                model,
                variation_count,
                temperature,
                api_key=api_key,
            )
        except AIBridgeException as e:
            raise PalmTextException(f"Error in generating AI data {e}")

    @classmethod
    def get_response(
        self,
        messages,
        context,
        examples,
        model="models/chat-bison-001",
        variation_count=1,
        temperature=0.5,
        api_key=None,
    ):
        try:
            api_key = api_key if api_key else parse_api_key("palm_api")
            palm.configure(api_key=api_key)
            response = palm.chat(
                model=model,
                context=context,
                examples=examples,
                messages=messages,
                candidate_count=variation_count,
            )
            message_value = {
                "items": {
                    "response": {
                        "messages": response.messages,
                        "candiates": response.candidates,
                    },
                    "created_at": time.time(),
                    "token_used": None,
                    "ai_service": "palm_chat",
                }
            }
            return message_value
        except AIBridgeException as e:
            raise PalmTextException(f"{e}")



from AIBridge.queue_integration.queue_model import ReddisMQ
from AIBridge.exceptions import AssignQueueException


class Assign:
    @classmethod
    def get_queue(self, mq_name):
        mq_obj = {"redis": ReddisMQ()}
        if mq_name not in mq_obj:
            raise AssignQueueException(
                f"No such queue with give name found ->{mq_name}"
            )
        return mq_obj[mq_name]



from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class AIResponse(Base):
    __tablename__ = "ai_reponse"
    id = Column(String, primary_key=True)
    model = Column(String)
    response = Column(String)
    updated_at = Column(Integer)
    created_at = Column(Integer)



from AIBridge.database.models.ai_response import AIResponse, Base
import json
import uuid
import calendar
import time
from AIBridge.exceptions import AIResponseException
from AIBridge.database.db_layer import DBLayer


class FetchAIResponse:
    @classmethod
    def get_time(self):
        return calendar.timegm(time.gmtime())

    @classmethod
    def save_response(self, response_data, model, id):
        table_exist = DBLayer.check_table(AIResponse, Base, method="save")
        res = {
            "id": id,
            "response": json.dumps(response_data),
            "model": model,
            "updated_at": self.get_time(),
            "created_at": self.get_time(),
        }
        data = DBLayer.save(AIResponse, Base, res)

    @classmethod
    def get_response(self, id):
        try:
            res = DBLayer.get_by_id(AIResponse, id)
        except:
            return {"messgae": "response in process"}
        return res



"""
Here we are fetching the prompt with reffrence to the id and make the complete prompt out of it,
We have vraibale and prompt data we dynamically bind the variable to the prompt and then we are returning the complete prompt
prompt become dynamic and independant of the input varibales whne the vraible changes no need to change the prompt

"""
from AIBridge.exceptions import (
    AIBridgeException,
    PromptSaveException,
    VaribalesException,
    PromptCompletionException,
)
from jinja2 import Environment
from AIBridge.database.models.prompts import Prompts
from AIBridge.database.models.variables import Variables
import json
from AIBridge.database.db_layer import DBLayer


class Completion:
    @classmethod
    def parser(self, prompt, json_data):
        try:
            template_env = Environment(cache_size=1000)
            template = template_env.from_string(prompt)
            parsed_string = template.render(json_data)
            return parsed_string
        except AIBridgeException as e:
            raise PromptCompletionException("Error in prasing the prompt")

    @classmethod
    def create_prompt_from_id(self, prompt_ids, prompt_data_list=[], variables_list=[]):
        prompts = []
        for index in range(len(prompt_ids)):
            prompt = DBLayer.get_by_id(Prompts, id=prompt_ids[index])
            prompt_data = (
                json.loads(prompt["prompt_data"])
                if not prompt_data_list
                else prompt_data_list[index]
            )
            variables = (
                json.loads(prompt["variables"])
                if not variables_list
                else variables_list[index]
            )
            for key, value in variables.items():
                variable = DBLayer.filter_table(Variables, **{"key": value})
                if variable is None:
                    raise VaribalesException(f"Invalid variable key->{key}")
                var_value = json.loads(variable["value"])
                prompt_data[key] = ",".join(map(str, var_value))
            prompt_string = self.parser(prompt["prompt"], prompt_data)
            prompts.append(prompt_string)
        return prompts

    @classmethod
    def create_prompt(self, prompt_list, prompt_data_list=[], variables_list=[]):
        prompts = []
        for index in range(len(prompt_list)):
            variables = {}
            prompt_string = prompt_list[index]
            if prompt_data_list:
                variables.update(prompt_data_list[index])
            if variables_list:
                for key, value in variables_list[index].items():
                    variable = DBLayer.filter_table(Variables, **{"key": value})
                    if variable is None:
                        raise VaribalesException(f"Invalid variable key->{key}")
                    var_value = json.loads(variable["value"])
                    variables[key] = ",".join(map(str, var_value))
            prompt_string = self.parser(prompt_string, variables)
            prompts.append(prompt_string)
        return prompts



from abc import ABC, abstractmethod
from AIBridge.database.table_oprations import create_table
from AIBridge.database.table_oprations import check_table_exist
from AIBridge.database.session import db_session
from AIBridge.exceptions import DatabaseException
from sqlalchemy.orm import class_mapper
import json
from AIBridge.setconfig import SetConfig
from sqlalchemy.orm import Query

config = SetConfig.read_yaml()

session = db_session()


class SQLProcess(ABC):
    @abstractmethod
    def save(table_obj, base, data):
        pass

    @abstractmethod
    def update(table_obj, data):
        pass

    @abstractmethod
    def get_by_id(table_obj, id):
        pass

    @abstractmethod
    def get_all(table_obj, page):
        pass

    @abstractmethod
    def delete(table_obj, id):
        pass

    @abstractmethod
    def filter_table(table_obj, **parameters):
        pass

    @abstractmethod
    def check_table(table_obj, base, method):
        pass


class SQL(SQLProcess):
    @classmethod
    def check_table(self, table_obj, base=None, method=""):
        if not check_table_exist(table_obj.__tablename__):
            if method == "save":
                create_table(base)
                return True
            else:
                return False
        return True

    @classmethod
    def row_to_json(self, row):
        mapper = class_mapper(row.__class__)
        columns = [col.key for col in mapper.columns]
        row_dict = {col: getattr(row, col) for col in columns}
        return row_dict

    def save(self, table_obj, base, data):
        self.check_table(table_obj, base, method="save")
        model_instance = table_obj(**data)
        session.add(model_instance)
        session.commit()
        session.close()
        return data

    def update(self, table_obj, data, id):
        if not self.check_table(table_obj):
            raise DatabaseException(
                "No record  found to update the table, create new records"
            )
        table = session.query(table_obj).filter_by(id=id).first()
        if not table:
            raise DatabaseException("No record  found to update the table")
        for key, value in data.items():
            if value:
                setattr(table, key, value)
        session.commit()
        session.close()
        return self.row_to_json(table)

    def get_by_id(self, table_obj, id):
        if not self.check_table(table_obj):
            raise DatabaseException("No record  found")
        table = session.query(table_obj).filter_by(id=id).first()
        if not table:
            raise DatabaseException(f"No record  found with id: {id}")
        session.close()
        return self.row_to_json(table)

    def get_all(self, table_obj, page):
        if page <= 0:
            page = 1
        per_page = config["per_page"] if "per_page" in config else 10
        query = session.query(table_obj)
        pagination = query.slice((page - 1) * per_page, page * per_page).all()
        rows_json = [self.row_to_json(row) for row in pagination]
        total = len(session.query(table_obj).all())
        total_pages = total // per_page
        pagination_status = {
            "items": rows_json,
            "has_next": len(pagination) > per_page and page < pagination.total_pages,
            "has_prev": page > 1,
            "total": total,
            "pages": total_pages,
            "current_page": page,
            "per_page": per_page,
        }
        session.close()
        return pagination_status

    def delete(self, table_obj, id):
        if not self.check_table(table_obj):
            raise DatabaseException("No record  found")
        table = session.query(table_obj).filter_by(id=id).first()
        if not table:
            raise DatabaseException("No record  found ")
        session.delete(table)
        session.close()
        return {"message": "Record deleted successfully"}

    def filter_table(self, table_obj, **parameters):
        if not self.check_table(table_obj):
            return None
        query = session.query(table_obj)
        for param, value in parameters.items():
            query = query.filter(getattr(table_obj, param) == value)
        filtered_obj = query.first()
        if not filtered_obj:
            return None
        session.close()
        return self.row_to_json(filtered_obj)



import json
import csv
import xmltodict
import io


class IntoJson:
    @classmethod
    def csv_to_json(self, csv_data):
        csv_data = csv_data.replace("\n", "")
        csv_list = csv_data.split(",")
        json_dict = {"data": []}
        new_dict = {}
        for key in csv_list:
            new_dict[key] = ""
        json_dict["data"].append(new_dict)
        return json_dict

    @classmethod
    def xml_to_json(self, xml_data):
        if """<?xml version="1.0" encoding="UTF-8"?>""" in xml_data:
            xml_data = xml_data.replace(
                """<?xml version="1.0" encoding="UTF-8"?>""", ""
            )
        json_data = xmltodict.parse(xml_data)
        new_json = {}
        for key, value in json_data.items():
            new_json[key] = [value]
            break
        return new_json


class FromJson:
    @classmethod
    def json_to_csv(self, json_dict):
        # Convert JSON to CSV
        json_data = json_dict["data"]
        csv_output = io.StringIO()
        csv_writer = csv.DictWriter(csv_output, fieldnames=json_data[0].keys())
        csv_writer.writeheader()
        csv_writer.writerows(json_data)
        # Get the CSV data as a string
        csv_data = csv_output.getvalue()
        return csv_data

    @classmethod
    def json_to_xml(self, json_data):
        root_element_name = "root"
        xml_data = xmltodict.unparse({root_element_name: json_data}, pretty=True)
        return xml_data



from abc import ABC, abstractmethod
import redis
import threading
import concurrent.futures
from AIBridge.ai_services.process_mq import ProcessMQ
from AIBridge.ai_services.ai_services_response import FetchAIResponse
import logging
from AIBridge.exceptions import MessageQueueException
import json
from AIBridge.setconfig import SetConfig

logging.basicConfig(level=logging.ERROR)
logger = logging.getLogger(__name__)

config = SetConfig.read_yaml()


class QueueService(ABC):
    @abstractmethod
    def enqueue(self, message):
        pass

    @abstractmethod
    def dequeue(self, message):
        pass

    @abstractmethod
    def background_process(self, consumer_name):
        pass

    @abstractmethod
    def local_backgorund_process(self, consumer_name):
        pass


class ReddisMQ(QueueService):
    def enqueue(self, message):
        redis_client = redis.Redis(
            host=config["redis_host"], port=config["redis_port"], db=0
        )
        redis_client.xadd(config["stream_name"], message)

    def dequeue(self):
        with concurrent.futures.ThreadPoolExecutor() as executor:
            for index in range(int(config["no_of_threads"])):
                consumer_name = f"consumer{index+1}"
                executor.submit(self.background_process, args=(consumer_name,))

    def background_process(self, consumer_name, local_p=False):
        redis_client = redis.Redis(
            host=config["redis_host"], port=config["redis_port"], db=0
        )
        group_info = redis_client.xinfo_groups(config["stream_name"])
        group_exists = any(
            group["name"] == config["group_name"] for group in group_info
        )
        if group_exists:
            redis_client.xgroup_create(
                config["stream_name"], config["group_name"], id="0", mkstream=True
            )
        while True:
            try:
                messages = redis_client.xreadgroup(
                    config["group_name"],
                    consumer_name,
                    {config["stream_name"]: ">"},
                    count=10,
                    block=0,
                )
                if not messages and local_p == True:
                    break
                for stream, message_list in messages:
                    for message in message_list:
                        message_id = message[0].decode("utf-8")
                        message_data = message[1]
                        data = "data".encode("utf-8")
                        if data not in message_data:
                            continue
                        message_data = json.loads(message_data[data].decode("utf-8"))
                        if "ai_service" not in message_data:
                            continue
                        service_obj, response_obj = ProcessMQ.get_process_mq(
                            message_data["ai_service"]
                        )
                        try:
                            data = response_obj.get_response(
                                service_obj=service_obj, message_data=message_data
                            )

                            FetchAIResponse.save_response(
                                data, message_data["ai_service"], message_data["id"]
                            )
                        except MessageQueueException as e:
                            logger.exception(f"Error in the message queue->{e}")
                            FetchAIResponse.save_response(
                                {"error": f"{e}"},
                                message_data["ai_service"],
                                message_data["id"],
                            )
                        redis_client.xack(
                            config["stream_name"], config["group_name"], message_id
                        )
            except MessageQueueException as e:
                logger.exception(f"Error in the message queue->{e}")
        return consumer_name

    def local_backgorund_process(self):
        consumer_name = f"consumer{1}"
        self.background_process(consumer_name, local_p=True)
        return consumer_name



from AIBridge.queue_integration.assign_queue import Assign
import threading
from AIBridge.setconfig import SetConfig

config = SetConfig.read_yaml()


class MessageQ:
    @classmethod
    def mq_enque(self, message):
        mq_obj = Assign.get_queue(config["message_queue"])
        mq_obj.enqueue(message)

    @classmethod
    def mq_deque(self):
        mq_obj = Assign.get_queue(config["message_queue"])
        threading.Thread(target=mq_obj.dequeue).start()
        return {"message": "thread started"}

    @classmethod
    def local_process(self):
        mq_obj = Assign.get_queue(config["message_queue"])
        consumer_thread = threading.Thread(target=mq_obj.local_backgorund_process)
        consumer_thread.start()
        consumer_thread.join()
        return {"message": "thread finished for local process"}



from abc import ABC, abstractmethod
import json


class Caller(ABC):
    @abstractmethod
    def get_response(self, service_obj, message_data):
        pass


class OpenAiRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            max_tokens=message_data["max_tokens"],
            temperature=message_data["temperature"],
            output_format=message_data["output_format"],
            format_structure=message_data["format_structure"],
            api_key=message_data["api_key"],
        )
        return data


class OpenAiImageRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            image_data=json.loads(message_data["image_data"]),
            mask_image=json.loads(message_data["mask_image"]),
            variation_count=message_data["variation_count"],
            size=message_data["size"],
            process_type=message_data["process_type"],
            api_key=message_data["api_key"],
            model=message_data["api_key"],
            quality=message_data["quality"],
        )
        return data


class PalmTextRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            max_tokens=message_data["max_tokens"],
            temperature=message_data["temperature"],
            output_format=message_data["output_format"],
            format_structure=message_data["format_structure"],
            api_key=message_data["api_key"],
        )
        return data


class PalmChatRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            messages=message_data["messages"],
            context=message_data["context"],
            examples=message_data["examples"],
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            temperature=message_data["temperature"],
            api_key=message_data["api_key"],
        )
        return data


class StableDuffusionRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            image_data=json.loads(message_data["image_data"]),
            mask_image=json.loads(message_data["mask_image"]),
            negative_prompt=message_data["negative_prompt"],
            width=message_data["width"],
            height=message_data["height"],
            samples=message_data["samples"],
            num_inference_steps=message_data["num_inference_steps"],
            safety_checker=message_data["safety_checker"],
            enhance_prompt=message_data["enhance_prompt"],
            guidance_scale=message_data["guidance_scale"],
            strength=message_data["strength"],
            seed=message_data["seed"],
            webhook=message_data["webhook"],
            track_id=message_data["track_id"],
            multi_lingual=message_data["multi_lingual"],
            panorama=message_data["panorama"],
            self_attention=message_data["self_attention"],
            upscale=message_data["upscale"],
            embeddings_model=message_data["embeddings_model"],
            action=message_data["action"],
            message_queue=message_data["message_queue"],
            api_key=message_data["api_key"],
        )
        return data


class CohereRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            max_tokens=message_data["max_tokens"],
            temperature=message_data["temperature"],
            output_format=message_data["output_format"],
            format_structure=message_data["format_structure"],
            api_key=message_data["api_key"],
        )
        return data


class JarasicTextRes(Caller):
    @classmethod
    def get_response(self, service_obj, message_data):
        data = service_obj.get_response(
            prompts=json.loads(message_data["prompts"]),
            model=message_data["model"],
            variation_count=message_data["variation_count"],
            max_tokens=message_data["max_tokens"],
            temperature=message_data["temperature"],
            output_format=message_data["output_format"],
            format_structure=message_data["format_structure"],
            api_key=message_data["api_key"],
        )
        return data



from AIBridge.prompts.prompt_completion import Completion
from AIBridge.ai_services.ai_abstraction import AIInterface
from AIBridge.exceptions import StableDiffusionException, AIBridgeException
import json
import uuid
from AIBridge.constant.constant import STABLE_DIFFUSION_TYPES
from AIBridge.constant.common import parse_api_key, check_url
import requests
import time


class StableDiffusion(AIInterface):
    @classmethod
    def generate(
        self,
        prompts: list[str] = [],
        prompt_ids: list[str] = [],
        prompt_data: list[dict] = [],
        variables: list[dict] = [],
        image_data: list[str] = [],
        mask_image: list[str] = [],
        negative_prompt: str = "",
        width: str = "512",
        height: str = "512",
        variation_count: int = 1,
        enhance_prompt: str = "Yes",
        num_inference_steps: str = "20",
        guidance_scale: float = 7.5,
        strength: float = 0.7,
        safety_checker: str = "yes",
        multi_lingual: str = "no",
        panorama: str = "no",
        self_attention: str = "no",
        upscale: str = "no",
        embeddings_model=None,
        webhook=None,
        track_id=None,
        seed=None,
        action="text2img",
        message_queue=False,
        api_key=None,
    ):
        try:
            if prompts and prompt_ids:
                raise StableDiffusionException(
                    "please provide either prompts or prompts ids at atime"
                )
            if not prompts and not prompt_ids:
                raise StableDiffusionException(
                    "Either provide prompts or prompts ids to genrate the data"
                )
            if action not in STABLE_DIFFUSION_TYPES:
                raise StableDiffusionException(
                    f"action should be one of the{STABLE_DIFFUSION_TYPES}"
                )
            if action == "image2image" or action == "inpaint":
                if not image_data:
                    raise StableDiffusionException(
                        "Please enter image link in image data"
                    )
            if action == "inpaint" and not mask_image:
                raise StableDiffusionException(
                    "Please enter mask image link in mask image"
                )
            if mask_image:
                if len(mask_image) != len(image_data):
                    raise StableDiffusionException(
                        "mask_image length should be equal to image_data length",
                    )
                check_url(image_data)
            if prompt_ids:
                prompts_list = Completion.create_prompt_from_id(
                    prompt_ids=prompt_ids,
                    prompt_data_list=prompt_data,
                    variables_list=variables,
                )
            if prompts:
                if prompt_data or variables:
                    prompts_list = Completion.create_prompt(
                        prompt_list=prompts,
                        prompt_data_list=prompt_data,
                        variables_list=variables,
                    )
                else:
                    prompts_list = prompts
            if image_data:
                if prompts_list:
                    if len(image_data) != len(prompts_list):
                        raise StableDiffusionException(
                            "image_data length should be equal to prompts length",
                        )
                check_url(image_data)
            if message_queue:
                id = uuid.uuid4()
                message_data = {
                    "id": str(id),
                    "prompts": json.dumps(prompts_list),
                    "negative_prompt": negative_prompt,
                    "ai_service": "stable_diffusion",
                    "image_data": json.dumps(image_data),
                    "mask_image": json.dumps(mask_image),
                    "action": action,
                    "width": width,
                    "height": height,
                    "samples": str(variation_count),
                    "num_inference_steps": num_inference_steps,
                    "safety_checker": safety_checker,
                    "enhance_prompt": enhance_prompt,
                    "guidance_scale": guidance_scale,
                    "strength": strength,
                    "seed": seed,
                    "webhook": webhook,
                    "track_id": track_id,
                    "multi_lingual": multi_lingual,
                    "panorama": panorama,
                    "self_attention": self_attention,
                    "upscale": upscale,
                    "embeddings_model": embeddings_model,
                    "message_queue": message_queue,
                    "api_key": api_key,
                }
                message = {"data": json.dumps(message_data)}
                from AIBridge.queue_integration.message_queue import MessageQ

                MessageQ.mq_enque(message=message)
                return {"response_id": str(id)}
            return self.get_response(
                prompts=prompts_list,
                image_data=image_data,
                mask_image=mask_image,
                negative_prompt=negative_prompt,
                width=width,
                height=height,
                samples=str(variation_count),
                num_inference_steps=num_inference_steps,
                safety_checker=safety_checker,
                enhance_prompt=enhance_prompt,
                guidance_scale=guidance_scale,
                strength=strength,
                seed=seed,
                webhook=webhook,
                track_id=track_id,
                multi_lingual=multi_lingual,
                panorama=panorama,
                self_attention=self_attention,
                upscale=upscale,
                embeddings_model=embeddings_model,
                action=action,
                message_queue=message_queue,
                api_key=api_key,
            )
        except AIBridgeException as e:
            raise StableDiffusionException(f"{e}")

    @classmethod
    def get_response(
        self,
        prompts,
        image_data=[],
        mask_image=[],
        negative_prompt: str = "",
        width: str = "512",
        height: str = "512",
        samples: str = "1",
        enhance_prompt: str = "Yes",
        num_inference_steps: str = "20",
        guidance_scale: float = 7.5,
        strength: float = 0.7,
        safety_checker: str = "yes",
        multi_lingual: str = "no",
        panorama: str = "no",
        self_attention: str = "no",
        upscale: str = "no",
        embeddings_model=None,
        webhook=None,
        track_id=None,
        seed=None,
        action="text2img",
        message_queue=False,
        api_key=None,
    ):
        api_key = api_key if api_key else parse_api_key("stable_diffusion")
        payload = {
            "key": api_key,
            "prompt": "",
            "negative_prompt": negative_prompt,
            "width": width,
            "height": height,
            "samples": samples,
            "num_inference_steps": num_inference_steps,
            "seed": seed,
            "guidance_scale": guidance_scale,
            "safety_checker": safety_checker,
            "multi_lingual": multi_lingual,
            "panorama": panorama,
            "self_attention": self_attention,
            "upscale": upscale,
            "embeddings_model": embeddings_model,
            " enhance_prompt": enhance_prompt,
            "strength": strength,
            "webhook": webhook,
            "track_id": track_id,
        }
        headers = {"Content-Type": "application/json"}
        url = "https://stablediffusionapi.com/api/v3/text2img"
        if action == "img2img":
            url = "https://stablediffusionapi.com/api/v3/img2img"
        if action == "inpaint":
            url = "https://stablediffusionapi.com/api/v3/inpaint"
        data = []
        for index, prompt in enumerate(prompts):
            payload["prompt"] = prompt
            if image_data:
                payload["init_image"] = image_data[index]
            if mask_image:
                payload["mask_image"] = mask_image[index]
            response = requests.post(url, headers=headers, data=json.dumps(payload))
            res = response.json()
            if not message_queue:
                if res["status"] == "error" or res["status"] == "failed":
                    raise StableDiffusionException(res["message"])
            _data = [json.dumps({"url": url}) for url in res["output"]]
            data.append({"data": _data})
        message_value = {
            "items": {
                "response": data,
                "token_used": 0,
                "created_at": time.time(),
                "ai_service": "stable_diffusion",
            }
        }
        return message_value

    @classmethod
    def fetch_image(self, fetch_id):
        api_key = parse_api_key("stable_diffusion")
        payload = {
            "key": api_key,
        }
        url = "https://stablediffusionapi.com/api/v3/fetch/{id}"
        url = url.format(id=fetch_id)
        headers = {"Content-Type": "application/json"}
        response = requests.post(url, headers=headers, data=json.dumps(payload))
        res = response.json()
        if res["status"] == "error" or res["status"] == "failed":
            raise StableDiffusionException(res["message"])
        return res

    @classmethod
    def system_load(self):
        api_key = parse_api_key("stable_diffusion")
        payload = {
            "key": api_key,
        }
        headers = {"Content-Type": "application/json"}
        url = "https://stablediffusionapi.com/api/v3/system_load"
        response = requests.post(url, headers=headers, data=json.dumps(payload))
        res = response.json()
        if "status" in res:
            if res["status"] == "error" or res["status"] == "failed":
                raise StableDiffusionException(res["message"])
        return res



from AIBridge.constant.common import get_database_obj

database_obj = get_database_obj()


class DBLayer:
    @classmethod
    def save(self, table_obj, base, data):
        return database_obj.save(table_obj, base, data)

    @classmethod
    def update(self, table_obj, data, id):
        return database_obj.update(table_obj, data, id)

    @classmethod
    def get_by_id(self, table_obj, id):
        return database_obj.get_by_id(table_obj, id)

    @classmethod
    def get_all(self, table_obj, page=1):
        return database_obj.get_all(table_obj, page)

    @classmethod
    def delete(self, table_obj, id):
        return database_obj.delete(table_obj, id)

    @classmethod
    def filter_table(self, table_obj, **filter):
        return database_obj.filter_table(table_obj, **filter)

    @classmethod
    def check_table(self, table_obj, base, method):
        return database_obj.check_table(table_obj, base, method)



import io
import os
import sys
from shutil import rmtree
from setuptools import find_packages, setup, Command

NAME = "aibridge_test"
DESCRIPTION = 'Bridge for LLM"s'
URL = "https://github.com/me/myproject"
EMAIL = "ashish.tilekar@opsfuse.com"
AUTHOR = "Ashish Tilekar"
REQUIRES_PYTHON = ">=3.9.0"
VERSION = "0.1.7"
REQUIRED = [
    "openai<=1.7.1",
    "SQLAlchemy>=2.0.19",
    "redis>=4.6.0",
    "PyYAML>=6.0.1",
    "Jinja2>=3.1.2",
    "pymongo>=4.4.1",
    "sqlparse>=0.4.4",
    "jsonschema>=4.18.4",
    "Pillow>=10.0.0",
    "google-generativeai>=0.1.0",
    "cohere>=4.24",
    "ai21>=1.2.8",
    "xmltodict>=0.13.0",
]

here = os.path.abspath(os.path.dirname(__file__))
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

about = {}
about["__version__"] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Publishing the AIBRIDGE1"
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=[]),
    install_requires=REQUIRED,
    include_package_data=True,
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    # $ setup.py publish support.
    cmdclass={
        "upload": UploadCommand,
    },
)



import time
import uuid
from AIBridge.exceptions import PalmTextException, AIBridgeException
from AIBridge.prompts.prompt_completion import Completion
from AIBridge.ai_services.ai_abstraction import AIInterface
from AIBridge.output_validation.active_validator import ActiveValidator
import json
from AIBridge.constant.common import parse_fromat, parse_api_key
import google.generativeai as palm
from AIBridge.output_validation.convertors import FromJson, IntoJson


class PalmText(AIInterface):
    """
    Base class for PalmText's Services
    """

    @classmethod
    def generate(
        self,
        prompts: list[str] = [],
        prompt_ids: list[str] = [],
        prompt_data: list[dict] = [],
        variables: list[dict] = [],
        output_format: list[str] = [],
        format_strcture: list[str] = [],
        model="models/text-bison-001",
        variation_count: int = 1,
        max_tokens: int = 10000,
        temperature: float = 0.5,
        message_queue=False,
        api_key=None,
        output_format_parse=True,
    ):
        try:
            if prompts and prompt_ids:
                raise PalmTextException(
                    "please provide either prompts or prompts ids at atime"
                )
            if not prompts and not prompt_ids:
                raise PalmTextException(
                    "Either provide prompts or prompts ids to genrate the data"
                )
            if prompt_ids:
                prompts_list = Completion.create_prompt_from_id(
                    prompt_ids=prompt_ids,
                    prompt_data_list=prompt_data,
                    variables_list=variables,
                )
            if prompts:
                if prompt_data or variables:
                    prompts_list = Completion.create_prompt(
                        prompt_list=prompts,
                        prompt_data_list=prompt_data,
                        variables_list=variables,
                    )
                else:
                    prompts_list = prompts
            if output_format:
                if len(output_format) != len(prompts_list):
                    raise AIBridgeException(
                        "length of output_format must be equal to length of the prompts"
                    )
            if format_strcture:
                if len(format_strcture) != len(prompts_list):
                    raise AIBridgeException(
                        "length of format_strcture must be equal to length of the prompts"
                    )
            updated_prompts = []
            for _prompt in prompts_list:
                format = None
                format_str = None
                if output_format:
                    format = output_format[prompts_list.index(_prompt)]
                if format_strcture:
                    format_str = format_strcture[prompts_list.index(_prompt)]
                if output_format_parse:
                    u_prompt = parse_fromat(
                        _prompt, format=format, format_structure=format_str
                    )
                    updated_prompts.append(u_prompt)
            if not updated_prompts:
                updated_prompts = prompts_list
            if message_queue:
                id = uuid.uuid4()
                message_data = {
                    "id": str(id),
                    "prompts": json.dumps(updated_prompts),
                    "model": model,
                    "variation_count": variation_count,
                    "max_tokens": max_tokens,
                    "temperature": temperature,
                    "ai_service": "palm_api",
                    "output_format": json.dumps(output_format),
                    "format_structure": json.dumps(format_strcture),
                    "api_key": api_key,
                }
                message = {"data": json.dumps(message_data)}
                from AIBridge.queue_integration.message_queue import MessageQ

                MessageQ.mq_enque(message=message)
                return {"response_id": str(id)}
            return self.get_response(
                updated_prompts,
                model,
                variation_count,
                max_tokens,
                temperature,
                output_format,
                format_strcture,
                api_key=api_key,
            )
        except AIBridgeException as e:
            raise PalmTextException(f"Error in generating AI data {e}")

    @classmethod
    def get_tokens(Self, model, text):
        tokens = palm.count_text_tokens(model=model, prompt=text)
        return tokens["token_count"]

    @classmethod
    def get_response(
        self,
        prompts,
        model="models/text-bison-001",
        variation_count=1,
        max_tokens=10000,
        temperature=0.5,
        output_format=[],
        format_structure=[],
        api_key=None,
    ):
        try:
            if output_format:
                if isinstance(output_format, str):
                    output_format = json.loads(output_format)
            if format_structure:
                if isinstance(format_structure, str):
                    format_structure = json.loads(format_structure)
            if not prompts:
                raise PalmTextException("No prompts provided")
            api_key = api_key if api_key else parse_api_key("palm_api")
            palm.configure(api_key=api_key)
            model_output = []
            token_used = 0
            for index, prompt in enumerate(prompts):
                response = palm.generate_text(
                    model=model,
                    prompt=prompt,
                    temperature=temperature,
                    max_output_tokens=max_tokens,
                    candidate_count=variation_count,
                )
                token_used = token_used + self.get_tokens(model, prompt)
                for res in response.candidates:
                    content = res["output"]
                    token_used = token_used + self.get_tokens(model, content)
                    if output_format:
                        _formatter = output_format[index]
                        try:
                            if _formatter == "csv":
                                content = FromJson.json_to_csv(json.loads(content))
                            elif _formatter == "xml":
                                content = FromJson.json_to_xml(json.loads(content))
                        except AIBridgeException as e:
                            raise PalmTextException(
                                f"Ai response is not in valid {_formatter}"
                            )
                        if _formatter == "json":
                            _validate_obj = ActiveValidator.get_active_validator(
                                _formatter
                            )
                            try:
                                content = _validate_obj.validate(
                                    content,
                                    schema=format_structure[prompts.index(prompt)]
                                    if format_structure
                                    else None,
                                )
                            except AIBridgeException as e:
                                content_error = {
                                    "error": f"{e}",
                                    "ai_response": content,
                                }
                                content = json.dumps(content_error)
                    if index >= len(model_output):
                        model_output.append({"data": [content]})
                    else:
                        model_output[index]["data"].append(content)
            message_value = {
                "items": {
                    "response": model_output,
                    "token_used": token_used,
                    "created_at": time.time(),
                    "ai_service": "palm_api",
                }
            }
            return message_value
        except AIBridgeException as e:
            raise PalmTextException(f"{e}")



PRIORITY = [
    "high",
    "medium",
    "equal",
    "low",
]
FORMATTER = ["json", "csv", "xml"]
AI_SERVICES = ["open_ai", "palm_api", "stable_diffusion", "cohere_api", "ai21_api"]
DB_TYPES = ["sql", "nosql"]
CUSTOM_CONFIG_PATH = "AIBRIDGE_CONFIG"
LINUX_MACOS_PATH = "/etc/aibridge/aibridge_config.yaml"
OPENAI_IMAGE_TYPE = ["create", "variation", "edit"]
OPENAI_IMAGE_SIZES = ["256x256", "512x512", "1024x1024"]
STABLE_DIFFUSION_TYPES = [
    "text2img",
    "img2img",
    "inpaint",
]
FUNCTION_CALL_FORMAT = {
    "name": "create_date",
    "description": "Generate output in given key and valid json formate",
    "parameters": {},
}



from PIL import Image
import requests
from urllib.parse import urlparse
from AIBridge.exceptions import ImageException, AIBridgeException
from io import BytesIO
import base64
import sys


class ImageOptimise:
    TARGET_SIZE = 4 * 1024 * 1024

    @classmethod
    def image_resize(self, image, buf):
        current_size = len(buf.getvalue())
        width = image.width
        height = image.height
        bytes_im = None
        print(current_size, self.TARGET_SIZE)
        while current_size > self.TARGET_SIZE:
            width = int(width * 0.75)
            height = int(height * 0.75)
            resized_image = image.resize((width, height), Image.Resampling.LANCZOS)
            temp_file = BytesIO()
            resized_image.save(temp_file, format="PNG", quality=80)
            print("Original image size:", sys.getsizeof(temp_file) / (1 << 20), "MB")
            current_size = len(temp_file.getvalue())
            bytes_im = temp_file.getvalue()
            temp_file.seek(0)
            temp_file.close
        if not bytes_im:
            temp_file = BytesIO()
            image.save(temp_file, format="PNG", quality=80)
            print("xxxxxxxxxxxxxxxxxx:", sys.getsizeof(temp_file) / (1 << 20), "MB")
            bytes_im = temp_file.getvalue()
            temp_file.seek(0)
            temp_file.close

        return bytes_im

    @classmethod
    def get_image(self, image_data):
        image_obj = []
        try:
            for index, image in enumerate(image_data):
                result = urlparse(image)
                if all([result.scheme, result.netloc]):
                    response = requests.get(image)
                    if response.status_code == 200:
                        image_data = BytesIO(response.content)
                        image = Image.open(image_data)
                elif "base64" in image:
                    image = image.split(",")[1]
                    image = BytesIO(base64.b64decode(image))
                    image = Image.open(image)
                else:
                    with open(image, "rb") as image_file:
                        binary_data = image_file.read()
                    image_buffer = BytesIO(binary_data)
                    image = Image.open(image_buffer)
                if image.mode != "RGBA":
                    image = image.convert("RGBA")
                image_obj.append(image)
            return image_obj
        except AIBridgeException as e:
            raise ImageException(f"Error in reading the the image {e}")

    @classmethod
    def get_bytes_io(self, image_data):
        image_obj = []
        for image in image_data:
            buf = BytesIO()
            image.save(buf, save_all=True, format="PNG", quality=100)
            buf_size = len(buf.getvalue())
            print("buffer image size:", sys.getsizeof(buf) / (1 << 20), "MB")
            if buf_size > self.TARGET_SIZE:
                bytes_im = self.image_resize(image, buf)
            else:
                bytes_im = buf.getvalue()
            buf.close()
            image_obj.append(bytes_im)
        return image_obj

    @classmethod
    def set_dimension(self, image_data, mask_data):
        for index, image in enumerate(image_data):
            if image.size != mask_data[index].size:
                mask = mask_data[index].resize(image.size)
                print(image.size)
                mask_data[index] = mask
                print(mask_data[index].size)
        return mask_data



from AIBridge.database.sql_service import SQLProcess
import pymongo
from AIBridge.setconfig import SetConfig
from AIBridge.exceptions import DatabaseException

config = SetConfig.read_yaml()


def mongo_session():
    client = pymongo.MongoClient(config["database_uri"])
    db_name = config["database_name"] if "database_name" in config else "aibridge"
    db = client[db_name]
    return db


class Mongodb(SQLProcess):
    @classmethod
    def check_table(self, table_obj, base=None, method=""):
        db = mongo_session()
        collection_names = db.list_collection_names()
        if method == "save":
            client = db[table_obj.__tablename__]
        if table_obj.__tablename__ not in collection_names:
            return False
        return True

    def save(self, table_obj, base, data):
        db = mongo_session()
        client = db[table_obj.__tablename__]
        result = client.insert_one(data)
        inserted_document = client.find_one({"id": data["id"]})
        return inserted_document

    def update(self, table_obj, data, id):
        if not self.check_table(table_obj):
            raise DatabaseException("No record found  in collection to update")
        db = mongo_session()
        client = db[table_obj.__tablename__]
        update_query = {"id": id}
        for key, value in data.items():
            if value == None:
                del data[key]
        new_values = {"$set": data}
        update_result = client.update_one(update_query, new_values)
        inserted_document = client.find_one({"id": id})
        return inserted_document

    def get_by_id(self, table_obj, id):
        if not self.check_table(table_obj):
            raise DatabaseException("No record found in colection")
        db = mongo_session()
        client = db[table_obj.__tablename__]
        inserted_document = client.find_one({"id": id})
        if not inserted_document:
            raise DatabaseException(f"No record found in colection with id {id}")
        return inserted_document

    def get_all(self, table_obj, page):
        if page <= 0:
            page = 1
        per_page = config["per_page"] if "per_page" in config else 10
        if not self.check_table(table_obj):
            raise DatabaseException("No record found in colection")
        db = mongo_session()
        client = db[table_obj.__tablename__]
        page = 1
        per_page = 10
        skip_count = (page - 1) * per_page
        pagination = client.find().skip(skip_count).limit(per_page)
        rows_json = [doc for doc in pagination]
        total = client.count_documents({})
        total_pages = (total + per_page - 1) // per_page
        pagination_status = {
            "items": rows_json,
            "has_next": skip_count + per_page < total,
            "has_prev": page > 1,
            "total": total,
            "pages": total_pages,
            "current_page": page,
            "per_page": per_page,
        }
        return pagination_status

    def delete(self, table_obj, id):
        if not self.check_table(table_obj):
            raise DatabaseException("No record found in colection")
        db = mongo_session()
        client = db[table_obj.__tablename__]
        delete_result = client.delete_one({"id": id})
        if delete_result.deleted_count == 0:
            raise DatabaseException(f"No record found in colection with id {id}")
        return {"message": "Record deleted successfully"}

    def filter_table(self, table_obj, **parameters):
        if not self.check_table(table_obj):
            return None
        db = mongo_session()
        client = db[table_obj.__tablename__]
        parameters = {key: value for key, value in parameters.items()}
        filtered_doc = client.find_one(parameters)
        if not filtered_doc:
            return None
        return filtered_doc



from abc import ABC, abstractmethod


class AIInterface(ABC):
    @abstractmethod
    def generate(
        self,
        prompts,
        prompt_ids,
        prompt_data,
        variables,
        model,
        variation_count,
        max_tokens,
        temperature,
    ):
        pass

    @abstractmethod
    def get_response(
        self,
        prompts,
        model,
        variation_count,
        max_tokens,
        temperature,
    ):
        pass



from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base
Base = declarative_base()
class Variables(Base):
    __tablename__ = 'variables'
    id = Column(String, primary_key=True)
    key=Column(String,unique=True)
    value=Column(String)
    updated_at = Column(Integer)
    created_at = Column(Integer)

